import logging
from openapi_client import SimulationsApi
from openapi_client.models import (
    ConstantWind,
    TimeseriesWind,
    WindDragCoefficient,
)
from .base import EventWrapper, InitialWrapper

logger = logging.getLogger(__name__)


class WindDragCoefficientWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = WindDragCoefficient
    api_path = "wind_drag_coefficient"
    scenario_name = "winddragcoefficient"


class ConstantWindWrapper(EventWrapper):
    api_class = SimulationsApi
    model = ConstantWind
    api_path: str = "wind_constant"
    scenario_name = "constantwind"


class TimeseriesWindWrapper(EventWrapper):
    api_class = SimulationsApi
    model = TimeseriesWind
    api_path: str = "wind_timeseries"
    scenario_name = "timeserieswind"


WRAPPERS = [
    WindDragCoefficientWrapper,
    ConstantWindWrapper,
    TimeseriesWindWrapper,
]



from rich.console import Console
from rich.theme import Theme


custom_theme = Theme(
    {
        "info": "black",
        "success": "bold spring_green4",
        "warning": "bold gold3",
        "error": "bold red3",
    }
)

console = Console(theme=custom_theme)



from .base import EventWrapper
from openapi_client import SimulationsApi
from openapi_client import ThreedimodelsApi
from typing import Dict
from openapi_client.models import RasterEdit


class RasterEditError(Exception):
    pass


class RasterEditWrapper(EventWrapper):
    api_class = SimulationsApi
    model = RasterEdit
    api_path: str = "raster_edits"
    scenario_name = "rasteredit"
    raster_type = None

    def initialize_instance(self, data: Dict):
        self.raster_type = data.pop("raster__type", None)
        if self.raster_type is None and "raster" not in data:
            raise RasterEditError(
                f"Please proved either `raster__type` or `raster`"
                f"the RasterEdit: {data}"
            )

        if self.raster_type is not None:
            # Try to lookup/find the raster.
            api = ThreedimodelsApi(self._api_client)
            res = api.threedimodels_rasters_list(
                self.simulation.threedimodel, type=self.raster_type
            )
            if res.count != 1:
                raise RasterEditError(
                    f"Could not find raster_type, raster_list "
                    f"response whas {res}"
                )

            # Set raster
            data["raster"] = res.results[0].id

        super().initialize_instance(data)


WRAPPERS = [RasterEditWrapper]



import logging
from pathlib import Path
from .base import EventWrapper
from typing import Dict
from openapi_client import SimulationsApi
from openapi_client.models import (
    ConstantSourcesSinks,
    TimeseriesSourcesSinks,
    LizardRasterSourcesSinks,
    LizardTimeseriesSourcesSinks,
    Upload,
    FileTimeseriesSourcesSinks,
    FileRasterSourcesSinks,
    NetCDFRasterSourcesSinks,
    NetCDFTimeseriesSourcesSinks,
)

from cmd_client.models.waitfor import WaitForProcessedFileWrapper

logger = logging.getLogger(__name__)


class ConstantSourcesSinksWrapper(EventWrapper):
    api_class = SimulationsApi
    model = ConstantSourcesSinks
    api_path: str = "sources_sinks_constant"
    scenario_name = "constantsourcessinks"


class SourcesSinksTimeseriesWrapper(EventWrapper):
    api_class = SimulationsApi
    model = TimeseriesSourcesSinks
    api_path: str = "sources_sinks_timeseries"
    scenario_name = "timeseriessourcessinks"


class SourcesSinksRasterLizardWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LizardRasterSourcesSinks
    api_path: str = "sources_sinks_rasters_lizard"
    scenario_name = "sourcessinksrasterlizard"


class SourcesSinksTimeseriesLizardWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LizardTimeseriesSourcesSinks
    api_path: str = "sources_sinks_timeseries_lizard"
    scenario_name = "sourcessinkstimeserieslizard"


class WaitForSourcesSinksTimeseriesFileWrapper(WaitForProcessedFileWrapper):
    api_class = SimulationsApi
    model = FileTimeseriesSourcesSinks
    scenario_name = "waitforsourcessinkstimeseriesfile"


class WaitForSourcesSinksTimeseriesNetCDFWrapper(
    WaitForSourcesSinksTimeseriesFileWrapper
):
    model = NetCDFTimeseriesSourcesSinks
    websocket_model_name = "NetCDFTimeseriesSourcesSinks"
    scenario_name = "waitforsourcessinkstimeseriesnetcdf"


class SourcesSinksTimeseriesNetCDFWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Upload
    api_path: str = "sources_sinks_timeseries_netcdf"
    scenario_name = "sourcessinkstimeseriesnetcdf"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 10,
        }
        wait_for_validation = WaitForSourcesSinksTimeseriesNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


class WaitForSourcesSinksRasterFileWrapper(WaitForProcessedFileWrapper):
    api_class = SimulationsApi
    model = FileRasterSourcesSinks
    scenario_name = "waitforsourcessinksrasterfile"


class WaitForSourcesSinksRasterNetCDFWrapper(
    WaitForSourcesSinksRasterFileWrapper
):
    model = NetCDFRasterSourcesSinks
    websocket_model_name = "NetCDFRasterSourcesSinks"
    scenario_name = "waitforsourcessinksrasternetcdf"


class SourcesSinksRasterNetCDFWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Upload
    api_path: str = "sources_sinks_rasters_netcdf"
    scenario_name = "sourcessinksrasternetcdf"
    filepath = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 10,
        }
        wait_for_validation = WaitForSourcesSinksRasterNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


WRAPPERS = [
    ConstantSourcesSinksWrapper,
    SourcesSinksTimeseriesWrapper,
    SourcesSinksRasterLizardWrapper,
    SourcesSinksTimeseriesLizardWrapper,
    SourcesSinksTimeseriesNetCDFWrapper,
    WaitForSourcesSinksTimeseriesNetCDFWrapper,
    SourcesSinksRasterNetCDFWrapper,
    WaitForSourcesSinksTimeseriesNetCDFWrapper,
]



from .actions import ActionWrapper
from .simulation import SimulationWrapper
from . import waitfor
from . import savedstate
from . import rain
from . import lateral
from . import sources_sinks
from . import wind
from . import breach
from . import rasteredit
from . import initial_waterlevels
from . import structure_control
from . import boundary_conditions

# Define a list of WRAPPERS
# A wrapper is an object that 'wraps' a certain
# resource of the openapi by automatically mapping
# all of it's openapi client methods.

# For every step in a scenario a wrapper is generated

WRAPPERS = (
    [
        ActionWrapper,
        SimulationWrapper,
    ]
    + rain.WRAPPERS
    + savedstate.WRAPPERS
    + waitfor.WRAPPERS
    + lateral.WRAPPERS
    + sources_sinks.WRAPPERS
    + wind.WRAPPERS
    + breach.WRAPPERS
    + rasteredit.WRAPPERS
    + initial_waterlevels.WRAPPERS
    + structure_control.WRAPPERS
    + boundary_conditions.WRAPPERS
)



from typing import Dict
from pathlib import Path

import logging
from openapi_client import SimulationsApi, UploadEventFile
from openapi_client.models import (
    ConstantLateral,
    TimeseriesLateral,
    FileLateral,
)
from .base import EventWrapper
from .waitfor import WaitForEventValidation

logger = logging.getLogger(__name__)


class WaitForConstantLateralWrapper(WaitForEventValidation):
    api_class = SimulationsApi
    model = ConstantLateral
    scenario_name = "waitforconstantlateral"


class WaitForTimeseriesLateralWrapper(WaitForEventValidation):
    api_class = SimulationsApi
    model = TimeseriesLateral
    scenario_name = "waitfortimeserieslateral"


class WaitForFileLateralWrapper(WaitForEventValidation):
    api_class = SimulationsApi
    model = FileLateral
    scenario_name = "waitforfilelateral"


class ConstantLateralWrapper(EventWrapper):
    api_class = SimulationsApi
    model = ConstantLateral
    api_path: str = "lateral_constant"
    scenario_name = "constantlateral"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForConstantLateralWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


class TimeseriesLateralWrapper(EventWrapper):
    api_class = SimulationsApi
    model = TimeseriesLateral
    api_path: str = "lateral_timeseries"
    scenario_name = "timeserieslateral"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForTimeseriesLateralWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


class FileLateralWrapper(EventWrapper):
    api_class = SimulationsApi
    model = UploadEventFile
    api_path: str = "lateral_file"
    scenario_name = "filelateral"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        file_laterals = self.api.simulations_events_lateral_file_list(
            simulation_pk=self.simulation.id
        )
        for file_lateral in file_laterals.results:
            if file_lateral.file:
                file_lateral.file.filename == self.instance.filename
                bulk_event_instance = file_lateral
                break
        assert bulk_event_instance is not None

        wait_for_validation = WaitForFileLateralWrapper(
            data=bulk_event_instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


WRAPPERS = [
    WaitForConstantLateralWrapper,
    WaitForTimeseriesLateralWrapper,
    WaitForFileLateralWrapper,
    ConstantLateralWrapper,
    TimeseriesLateralWrapper,
    FileLateralWrapper,
]



import logging
from typing import Dict
from openapi_client import SimulationsApi
from openapi_client import ThreedimodelsApi
from openapi_client.models import (
    OneDWaterLevel,
    TwoDWaterLevel,
    GroundWaterLevel,
    OneDWaterLevelPredefined,
    TwoDWaterRaster,
    GroundWaterRaster,
)
from .base import InitialWrapper

logger = logging.getLogger(__name__)


class OneDWaterLevelWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = OneDWaterLevel
    base_path = "simulations_initial"
    api_path = "1d_water_level_constant"
    scenario_name = "1dwaterlevelconstant"


class OneDWaterLevelPredefinedWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = OneDWaterLevelPredefined
    base_path = "simulations_initial"
    api_path = "1d_water_level_predefined"
    scenario_name = "1dwaterlevelpredefined"


class TwoDWaterLevelWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = TwoDWaterLevel
    base_path = "simulations_initial"
    api_path = "2d_water_level_constant"
    scenario_name = "2dwaterlevelconstant"


class TwoDWaterLevelRasterError(Exception):
    pass


class TwoDWaterLevelRasterWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = TwoDWaterRaster
    base_path = "simulations_initial"
    api_path = "2d_water_level_raster"
    scenario_name = "2dwaterlevelraster"
    lookup: str = None

    # initial_waterlevel__source_raster__type
    def initialize_instance(self, data: Dict):
        self.lookup = data.pop("initial_waterlevel__source_raster__type", None)
        if self.lookup is None and "initial_waterlevel" not in data:
            raise TwoDWaterLevelRasterError(
                f"Please proved either "
                f"`initial_waterlevel__source_raster__type` "
                f"or `initial_waterlevel`"
                f"for the TwoDWaterLevelRaster: {data}"
            )

        if self.lookup is not None:
            # Try to lookup/find the raster.
            api = ThreedimodelsApi(self._api_client)
            res = api.threedimodels_rasters_list(
                self.simulation.threedimodel, type=self.lookup
            )
            if res.count != 1:
                raise TwoDWaterLevelRasterError(
                    f"Could not find "
                    f"initial_waterlevel__source_raster__type, raster_list "
                    f"response whas {res}"
                )

            raster = res.results[0]

            # Now find the initial waterlevel related to this
            # raster
            res = api.threedimodels_initial_waterlevels_list(
                self.simulation.threedimodel
            )

            # Search through initial waterlevel options
            # for the correct one
            found = [
                x
                for x in res.results
                if int(x.source_raster_id) == int(raster.id)
            ]

            if len(found) != 1:
                raise TwoDWaterLevelRasterError(
                    f"Could not find "
                    f"initial_waterlevel__source_raster__type, "
                    f"initial_waterlevels list"
                    f"response was {res}"
                )

            # Set initial_waterlevel
            data["initial_waterlevel"] = res.results[0].id

        super().initialize_instance(data)


class GroundWaterLevelWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = GroundWaterLevel
    base_path = "simulations_initial"
    api_path = "groundwater_level_constant"
    scenario_name = "groundwaterlevelconstant"


class GroundWaterLevelRasterWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = GroundWaterRaster
    base_path = "simulations_initial"
    api_path = "groundwater_level_raster"
    scenario_name = "groundwaterlevelraster"


WRAPPERS = [
    OneDWaterLevelWrapper,
    OneDWaterLevelPredefinedWrapper,
    TwoDWaterLevelWrapper,
    TwoDWaterLevelRasterWrapper,
    GroundWaterLevelWrapper,
    GroundWaterLevelRasterWrapper,
]



from openapi_client import SimulationsApi
from openapi_client.models import Breach
from .base import EventWrapper


class BreachWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Breach
    api_path: str = "breaches"
    scenario_name = "breach"


WRAPPERS = [BreachWrapper]



import asyncio
import click
from datetime import datetime
from cmd_client.console import console
from cmd_client.models.monitor import ActiveSimulations

from cmd_client.commands.settings import (
    Settings,
    EndpointOption,
    refresh_api_key,
    get_settings,
)


async def monitor_active_simulations(endpoint):
    """
    runs the ActiveSimlations "run-run_monitor" task in the background
    """
    active_simulations = ActiveSimulations(endpoint)
    result = await asyncio.gather(
        active_simulations.run_monitor(), return_exceptions=True
    )
    if result:
        console.print(result, style="error")


@click.command(help="Show currently running simulations")
@click.option(
    "--endpoint",
    type=click.Choice(EndpointOption.__members__.keys()),
    default=EndpointOption.localhost.name,
    help="The endpoint where commands are run against",
)
@click.pass_context
def cli(ctx: click.Context, endpoint: str):
    settings = get_settings(endpoint)
    ctx.obj = settings
    refresh_api_key(settings.api_client.configuration)
    ctx.call_on_close(Settings.save_settings)
    start_time = datetime.now().strftime("%Y-%m-%d %H:%M")
    try:
        console.print(f"[{start_time}] Starting active simulations worker")
        asyncio.run(monitor_active_simulations(endpoint))
    except KeyboardInterrupt:
        pass
    else:
        console.print(
            f":sparkles: Bye bye, hope to see you soon! :sparkles:",
            style="success",
        )


if __name__ == "__main__":
    cli()



import logging
from .base import SimulationChildWrapper
from openapi_client import SimulationsApi
from openapi_client.models import TimedSavedStateUpdate, InitialSavedState
from .base import InitialWrapper


logger = logging.getLogger(__name__)


class TimedSavedStateWrapper(SimulationChildWrapper):
    api_class = SimulationsApi
    model = TimedSavedStateUpdate
    api_path: str = "create_saved_states_timed"
    scenario_name = "timedsavedstate"


class InitialSavedStateWrapper(InitialWrapper):
    api_class = SimulationsApi
    model = InitialSavedState
    api_path: str = "saved_state"
    scenario_name = "initialsavedstate"


WRAPPERS = [TimedSavedStateWrapper, InitialSavedStateWrapper]



from __future__ import annotations
from dataclasses import dataclass, field, asdict
from datetime import timedelta
from datetime import datetime
from enum import Enum
from pathlib import Path
import sys
from urllib.parse import urlparse
from functools import lru_cache

import click
import jwt
import yaml
from rich.prompt import Prompt

from openapi_client import Configuration, ApiClient, ApiException
from threedi_api_client.threedi_api_client import ThreediApiClient
from openapi_client.api.auth_api import AuthApi
from threedi_api_client.threedi_api_client import (
    is_token_usable,
    get_auth_token,
)

from cmd_client.console import console
from cmd_client.errors import ExitCodes

SCENARIO_DIR = Path(__file__).parent.parent / "scenarios"
REFRESH_TIME_DELTA = timedelta(hours=4).total_seconds()
EXPIRE_LEEWAY = -300


class EndpointOption(Enum):
    localhost = "http://localhost:8000/v3.0"
    staging = "https://api.staging.3di.live/v3.0"
    production = "https://api.3di.live/v3.0"


@lru_cache()
def get_settings(endpoint):
    return Settings(endpoint=endpoint)


@dataclass
class WebSocketSettings:
    api_base_url: str
    token: str
    host: str = field(init=False)
    proto: str = field(init=False)
    api_version: str = field(init=False)

    def __post_init__(self):
        parsed_url = urlparse(self.api_base_url)
        self.host = parsed_url.netloc
        self.proto = "wss" if parsed_url.scheme == "https" else "ws"
        self.api_version = parsed_url.path.lstrip("/")
        self.token = (
            self.token
            if self.token.startswith("Bearer")
            else f"Bearer {self.token}"
        )


@dataclass
class CachedConfig:
    username: str = ""
    access: str = ""
    refresh: str = ""
    organisation_uuid: str = ""
    result_folder: str = ""

    @classmethod
    def load_from_file(cls, file: Path) -> CachedConfig:
        """
        Loads saved settings from 3di_config.yaml. See save method for available attributes.
        """
        try:
            with open(file, "r") as f:
                cached_data = yaml.load(f, Loader=yaml.FullLoader)
                return cls(**cached_data)
        except OSError:
            # settings file does not yet exist
            return cls()


class Settings:
    """Settings that are saved (authentication) values between calls to the cli.

    The default settings are saved in the file '3di_config.yaml' which lives in
    the configuration folder of your operating system.
    """

    app_name = "3di-scenario"
    file_name = "3di_config.yaml"

    def __init__(self, endpoint: str = ""):
        self.endpoint = EndpointOption[endpoint].value
        self.cached_config = CachedConfig.load_from_file(self.config_file)
        self.websocket_settings = WebSocketSettings(
            api_base_url=self.endpoint, token=self.access
        )

    @property
    def username(self):
        return self.cached_config.username

    @property
    def access(self):
        return self.cached_config.access

    @property
    def refresh(self):
        return self.cached_config.refresh

    @property
    def organisation_uuid(self):
        return self.cached_config.organisation_uuid

    @property
    def result_folder(self):
        return self.cached_config.result_folder

    @property
    def scenario_dir(self):
        return SCENARIO_DIR

    @property
    def app_dir(self) -> Path:
        _app_dir = Path(click.get_app_dir(app_name=self.app_name))
        _app_dir.mkdir(parents=True, exist_ok=True)
        return _app_dir

    @property
    def config_file(self) -> Path:
        return self.app_dir / self.file_name

    @property
    def configuration(self):
        configuration = Configuration(
            host=self.endpoint,
            username=f"{self.username}",
            api_key={
                "Authorization": f"{self.access}",
                "refresh": f"{self.refresh}",
            },
            api_key_prefix={"Authorization": "Bearer"},
        )
        configuration.refresh_api_key_hook = refresh_api_key
        return configuration

    @property
    def api_client(self) -> ThreediApiClient:
        client = ApiClient(self.configuration)
        return client

    @property
    def scenarios(self):
        scenario_list = []
        scenario_files = self.scenario_dir.glob("*.yaml")
        for f in scenario_files:
            with f.open("r") as y:
                content = yaml.load_all(y, Loader=yaml.FullLoader)
                try:
                    meta = next(content)["meta"]
                    meta["file"] = f
                    scenario_list.append(meta)
                except KeyError:
                    raise AttributeError(
                        f"The scenario definition is missing the 'meta' section"
                    )
        return scenario_list

    def save(self):
        with open(self.config_file, "w") as file:
            yaml.dump(
                asdict(self.cached_config),
                default_flow_style=False,
                stream=file,
            )

    @staticmethod
    def save_settings():
        ctx = click.get_current_context()
        ctx.obj.save()


def do_refresh(token) -> bool:
    try:
        details = jwt.decode(
            token,
            options={"verify_signature": False},
            leeway=EXPIRE_LEEWAY,
        )
    except (jwt.exceptions.ExpiredSignatureError, jwt.exceptions.DecodeError):
        return True
    expiry_dt = datetime.fromtimestamp(details["exp"])
    sec_left = (expiry_dt - datetime.utcnow()).total_seconds()
    return sec_left <= REFRESH_TIME_DELTA


def refresh_api_key(config: Configuration):
    """Refreshes the access key if its expired"""

    api_key = config.api_key.get("Authorization")
    if not do_refresh(api_key) and is_token_usable(api_key):
        return

    refresh_key = config.api_key.get("refresh")
    if is_token_usable(refresh_key):
        api_client = ApiClient(Configuration(config.host))
        auth = AuthApi(api_client)
        token = auth.auth_refresh_token_create(
            {"refresh": config.api_key["refresh"]}
        )
    else:
        console.rule(
            f":key: Authentication required for {config.host}",
            characters="*",
            style="gold3",
        )
        settings: Settings = click.get_current_context().obj
        kwargs = {}
        try:
            username = settings.username
            if username:
                kwargs.update({"default": username})
        except AttributeError:
            pass
        username = Prompt.ask("Username", **kwargs)
        password = Prompt.ask("Password", password=True)

        try:
            token = get_auth_token(username, password, config.host)
        except ValueError as err:
            console.print(f":collision: {err}", style="error")
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)

        # Update the settings
        settings.cached_config.username = username
        settings.cached_config.access = token.access
        settings.cached_config.refresh = token.refresh

    config.api_key = {"Authorization": token.access, "refresh": token.refresh}



import logging
from pathlib import Path
from typing import Dict
from openapi_client import SimulationsApi
from openapi_client.models import (
    ConstantRain,
    TimeseriesRain,
    LizardRasterRain,
    LizardTimeseriesRain,
    Upload,
    FileTimeseriesRain,
    FileRasterRain,
    NetCDFRasterRain,
    NetCDFTimeseriesRain,
    LocalRain,
)

from cmd_client.models.base import EventWrapper
from cmd_client.models.waitfor import WaitForProcessedFileWrapper

logger = logging.getLogger(__name__)


class ConstantRainWrapper(EventWrapper):
    api_class = SimulationsApi
    model = ConstantRain
    api_path: str = "rain_constant"
    scenario_name = "constantrain"


class LocalConstantRainWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LocalRain
    api_path: str = "rain_local_constant"
    scenario_name = "localrainconstant"


class RainTimeseriesWrapper(EventWrapper):
    api_class = SimulationsApi
    model = TimeseriesRain
    api_path: str = "rain_timeseries"
    scenario_name = "timeseriesrain"


class LocalRainTimeseriesWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LocalRain
    api_path: str = "rain_local_timeseries"
    scenario_name = "localraintimeseries"


class RainRasterLizardWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LizardRasterRain
    api_path: str = "rain_rasters_lizard"
    scenario_name = "rainrasterlizard"


class RainTimeseriesLizardWrapper(EventWrapper):
    api_class = SimulationsApi
    model = LizardTimeseriesRain
    api_path: str = "rain_timeseries_lizard"
    scenario_name = "raintimeserieslizard"


class WaitForProcessedTimeseriesFileWrapper(WaitForProcessedFileWrapper):
    api_class = SimulationsApi
    model = FileTimeseriesRain
    scenario_name = "waitforraintimeseriesfile"


class WaitForRainTimeseriesNetCDFWrapper(
    WaitForProcessedTimeseriesFileWrapper
):
    model = NetCDFTimeseriesRain
    websocket_model_name = "NetCDFTimeseriesRain"
    scenario_name = "waitforraintimeseriesnetcdf"


class RainTimeseriesNetCDFWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Upload
    api_path: str = "rain_timeseries_netcdf"
    scenario_name = "raintimeseriesnetcdf"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 10,
        }
        wait_for_validation = WaitForRainTimeseriesNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


class WaitForProcessedRasterFileWrapper(WaitForProcessedFileWrapper):
    api_class = SimulationsApi
    model = FileRasterRain
    scenario_name = "waitforrainrasterfile"


class WaitForRainRasterNetCDFWrapper(WaitForProcessedRasterFileWrapper):
    model = NetCDFRasterRain
    websocket_model_name = "NetCDFRasterRain"
    scenario_name = "waitforrainrasternetcdf"


class RainRasterNetCDFWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Upload
    api_path: str = "rain_rasters_netcdf"
    scenario_name = "rainrasternetcdf"
    filepath = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 10,
        }
        wait_for_validation = WaitForRainRasterNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


WRAPPERS = [
    ConstantRainWrapper,
    RainTimeseriesWrapper,
    RainRasterLizardWrapper,
    RainTimeseriesLizardWrapper,
    RainTimeseriesNetCDFWrapper,
    WaitForRainTimeseriesNetCDFWrapper,
    RainRasterNetCDFWrapper,
    WaitForRainTimeseriesNetCDFWrapper,
    LocalConstantRainWrapper,
    LocalRainTimeseriesWrapper,
]



import asyncio
import logging
from uuid import uuid4
from datetime import datetime
from threedi_api_client.threedi_api_client import ThreediApiClient
from cmd_client.parser import ScenarioParser

logging.basicConfig(level=logging.INFO)


async def main():
    organisation = "61f5a464c35044c19bc7d4b42d7f58cb"
    threedimodel_id = 14
    simulation_name = "run_" + uuid4().hex

    context = {
        "threedimodel_id": threedimodel_id,
        "organisation_uuid": organisation,
        "simulation_name": simulation_name,
        "datetime_now": datetime.utcnow().isoformat(),
    }

    file_path = "scenarios/livesite_constant_rain_default.yaml"
    file_path = "scenarios/livesite_rain_timeseries.yaml"
    file_path = "scenarios/livesite_lateral.yaml"
    file_path = "scenarios/start_shutdown.yaml"
    # file_path = 'scenarios/start_pause.yaml'

    config = {
        "API_HOST": "http://localhost:8000/v3.0",
        "API_USERNAME": "root",
        "API_PASSWORD": "root2",
    }

    client = ThreediApiClient(config=config)
    parser = ScenarioParser(file_path, context)
    scenario = parser.parse(client)

    # Save simulation before using it in the websocket
    scenario.simulation.save()

    await scenario.execute()


if __name__ == "__main__":
    asyncio.run(main())



from concurrent import futures
from urllib.parse import urlparse
from pathlib import Path
from typing import Iterable

import requests
from openapi_client import Download
from cmd_client.console import console
from rich.progress import (
    BarColumn,
    DownloadColumn,
    TextColumn,
    TransferSpeedColumn,
    TimeRemainingColumn,
    Progress,
    TaskID,
)

CHUNK_SIZE = 16 * 1024

progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
    console=console,
)


def download_worker(task_id: TaskID, url: str, file_path: Path) -> None:
    with requests.get(url, stream=True) as r:
        with open(file_path, "wb") as f:
            progress.start_task(task_id)
            for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                f.write(chunk)
                progress.update(task_id, advance=len(chunk))


def download_files(downloads: Iterable[Download], output_path: Path) -> bool:
    jobs = []
    success = True
    with progress:
        with futures.ThreadPoolExecutor(max_workers=4) as pool:
            for download in downloads:
                url = urlparse(download.get_url)
                file_name = Path(url.path).name
                file_path = output_path / file_name
                task_id = progress.add_task(
                    "download",
                    filename=file_name,
                    start=False,
                    total=download.size,
                )
                jobs.append(
                    pool.submit(
                        download_worker, task_id, download.get_url, file_path
                    )
                )
        for job in futures.as_completed(jobs):
            try:
                job.result()
            except requests.RequestException as err:
                progress.console.print(err, style="error")
                success = False
    return success



import asyncio
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
import json
from typing import List
from pathlib import Path
import sys

import click
from openapi_client.api import ThreedimodelsApi
from threedi_api_client.threedi_api_client import ThreediApiClient


import yaml

from cmd_client.commands.slack_notify import SlackNotifier
from cmd_client.models.scenario import FailedStep
from cmd_client.parser import ScenarioParser
from cmd_client.models.scenario import Scenario
from cmd_client.console import console
from cmd_client.errors import ExitCodes
from cmd_client.errors import LoadCredentialsError
from cmd_client.errors import LoadSuiteError
from cmd_client.commands.settings import WebSocketSettings


from rich.table import Table
from rich.padding import Padding
from rich import box


class SuiteError(BaseException):
    pass


@dataclass
class SlackNotify:
    channel: str
    crashes: bool = True
    all_success: bool = True


@dataclass
class User:
    username: str
    password: str

    def __repr__(self):
        return f"User(username='{self.username}', password='********')"


@dataclass
class Credentials:
    users: List[User]
    slack_token: str = None

    @classmethod
    def load_from_file(cls, file: Path = None):
        try:
            with file.open("r") as f:
                data = yaml.load(f, Loader=yaml.FullLoader) or {}
        except OSError:
            # settings file does not yet exist
            msg = f"Could not load credentials: {file}"
            raise LoadCredentialsError(msg)
        data["credentials"]["users"] = [
            User(**x) for x in data["credentials"].get("users", [])
        ]
        return Credentials(**data["credentials"])

    def get_user(self, username: str):
        found = [x for x in self.users if x.username == username]
        assert (
            len(found) == 1
        ), f"Did not find exactly one user for {username} in the credentials"

        return found[0]


@dataclass
class SuiteScenario:
    name: str
    scenario: Scenario
    finished: bool = False
    success: bool = False
    reason: str = "scenario is not finished"
    simulation_id = None


@dataclass
class Report:
    endpoint: str
    scenarios: List[SuiteScenario]

    def simulation_url(self, suite_scenario):
        return (
            f"<{self.endpoint}/simulations/"
            + f"{suite_scenario.simulation_id}|{suite_scenario.simulation_id}>"
        )

    @property
    def title(self):
        return f"Results framework suite run on {self.endpoint}"

    def get_slack_report(self):
        report_lines = [f"{self.endpoint}:"]
        for suite_scenario in self.scenarios:
            if suite_scenario.success:
                report_lines.append(
                    f':heavy_check_mark:   "{suite_scenario.name}" '
                    f"({self.simulation_url(suite_scenario)})"
                )
            else:
                report_lines.append(
                    f':x:   "{suite_scenario.name} "'
                    f"({self.simulation_url(suite_scenario)})"
                    f" errored "
                    f"due to: {suite_scenario.reason} "
                )
        return report_lines

    def print_console_report(self):
        console.print(Padding("", (1, 0)))
        console.rule(self.title, characters="•", style="black")
        console.print(Padding("", (1, 0)))
        grid = Table(
            show_header=True,
            box=box.SIMPLE,
            show_lines=False,
            width=console.width,
        )
        grid.add_column("Name")
        grid.add_column("URL", justify="left")
        grid.add_column("Details", justify="left")
        grid.add_column("Status", justify="right")
        for suite_scenario in self.scenarios:
            if suite_scenario.success:
                grid.add_row(
                    f"{suite_scenario.name}",
                    f"{self.endpoint}/simulations/{suite_scenario.simulation_id}",
                    "",
                    "[bold magenta]COMPLETED [green]:heavy_check_mark:",
                )
            else:
                grid.add_row(
                    f"{suite_scenario.name}",
                    f"{self.endpoint}/simulations/{suite_scenario.simulation_id}",
                    f"[bold gold3]{suite_scenario.reason}",
                    "[bold magenta]FAILED [red] :x:",
                )
        console.print(grid)


@dataclass
class Suite:
    endpoint: str
    path: str
    credentials: Credentials
    organisation_uuid: str
    # override_sessions: bool
    # sessions: int
    scenarios: List[SuiteScenario]
    user: str = None
    slack_notify: SlackNotify = None
    _resolved_models = None
    report: Report = field(init=False)

    def __post_init__(self):
        self.report = Report(endpoint=self.endpoint, scenarios=self.scenarios)

    @classmethod
    def resolve_model(cls, scenario_name, model_api, values, resolved_models):
        model = None

        if "threedimodel_filter" in values:
            cache_key = json.dumps(values["threedimodel_filter"])

            if cache_key in resolved_models:
                model = resolved_models[cache_key]
            else:
                res = model_api.threedimodels_list(
                    inp_success=True,
                    disabled=False,
                    **values["threedimodel_filter"],
                )
                if res.count == 0:
                    print(
                        f"Error: no models found for "
                        f"scenario: {scenario_name}"
                    )
                    return None

                if res.count > 0:
                    console.print(
                        f"! multiple models found for lookup: {values['threedimodel_filter']}",
                        style="warning",
                    )
                model = res.results[0].id

                resolved_models[cache_key] = model

        elif "threedimodel_id" in values:
            model = values["threedimodel_id"]
        else:
            print(f"No model defined for scenario: {scenario_name}")
        return model

    @classmethod
    def load_from_file(cls, file: Path = None):
        """
        :raises LoadCredentialsError if the credentials file could not be found or loaded
        :raises TypeError if the `file` argument is not of type `Path`
        :raises LoadSuiteError if the suite could not be loaded
        """
        console.rule(f"Loading suite {file.stem}", style="bold blue")

        if not isinstance(file, Path):
            raise TypeError(f"{file} must be of type pathlib.Path")
        try:
            with open(file, "r") as f:
                data = yaml.load(f, Loader=yaml.FullLoader) or {}
        except OSError:
            msg = f"Could not load suite: {file}"
            raise LoadSuiteError(msg)
        suite = data["suite"]
        base_path = file.parent

        if "slack_notify" in suite:
            suite["slack_notify"] = SlackNotify(**suite["slack_notify"])
            console.print(
                f"Suite results will be published in slack channel {suite['slack_notify'].channel}"
            )
        if "credentials" in suite:
            suite["credentials"] = Credentials.load_from_file(
                base_path / Path(suite["credentials"])
            )
        else:
            msg = "Could not find/load credentials file"
            raise LoadCredentialsError(msg)

        resolved_models = {}

        scenarios = []
        for scenario in suite.get("scenarios", []):
            for key, values in scenario.items():

                user = values.get("user", suite.get("user", None))

                if user is None:
                    if len(suite["credentials"].users) != 1:
                        msg = "No default user specified and multiple user options in credentials"
                        raise LoadCredentialsError(msg)
                    user = suite["credentials"].users[0]
                else:
                    user = suite["credentials"].get_user(user)

                config = {
                    "API_HOST": suite["endpoint"],
                    "API_USERNAME": user.username,
                    "API_PASSWORD": user.password,
                }

                client = ThreediApiClient(config=config)
                token = client.configuration.get_api_key_with_prefix(
                    "Authorization"
                )

                model_api = ThreedimodelsApi(client)

                model = cls.resolve_model(
                    key, model_api, values, resolved_models
                )

                # Try scenario.organisation__uuid, else suite organisation_uuid
                organisation_uuid = values.get(
                    "organisation_uuid", suite["organisation_uuid"]
                )

                context = {
                    "threedimodel_id": model,
                    "organisation_uuid": organisation_uuid,
                    "simulation_name": key,
                    "datetime_now": datetime.utcnow().isoformat(),
                }

                scenario_to_run = base_path / Path(values["file"])
                websocket_settings = WebSocketSettings(
                    api_base_url=suite["endpoint"], token=token
                )
                if model is not None:
                    parser = ScenarioParser(scenario_to_run, context)
                    scenario = parser.parse(
                        client, websocket_settings, base_path=base_path
                    )
                else:
                    scenario = None

                suite_scenario = SuiteScenario(name=key, scenario=scenario)

                if model is None:
                    suite_scenario.success = False
                    suite_scenario.reason = f"could not find the model"
                    suite_scenario.finished = True

                scenarios.append(suite_scenario)
        suite["scenarios"] = scenarios
        suite["path"] = base_path
        return Suite(**suite)


async def main(suite: Path):
    """runs the entire suite"""

    try:
        suite = Suite.load_from_file(suite)
    except (TypeError, LoadCredentialsError, LoadSuiteError) as err:
        console.print(err, style="error")
        sys.exit(ExitCodes.LOADING_SUITE_ERROR.value)

    for suite_scenario in suite.scenarios:
        if suite_scenario.finished:
            continue
        console.rule("Loading scenario...", style="bold blue")
        scenario = suite_scenario.scenario
        scenario.simulation.save()

        simulation_id = scenario.simulation.instance.id
        suite_scenario.simulation_id = simulation_id
        console.rule(f"Starting scenario run...", style="bold blue")
        try:
            await scenario.execute()
            suite_scenario.success = True
            suite_scenario.finished = True
            console.print(
                f":sparkles: Successfully finished scenario run :sparkles:",
                style="success",
            )
        except KeyboardInterrupt:
            pass
        except FailedStep as e:
            console.print(
                ":collision: Error running sim: {str(e)}", style="error"
            )
            suite_scenario.success = False
            suite_scenario.finished = True
            suite_scenario.reason = str(e)
        except Exception as e:
            console.print(
                ":collision: Error running sim: {str(e)}", style="error"
            )
            suite_scenario.success = False
            suite_scenario.finished = True
            suite_scenario.reason = str(e)

    # Print report
    suite.report.print_console_report()
    if suite.slack_notify:
        notifier = SlackNotifier(suite.credentials.slack_token)
        notifier.send_message(
            "\n".join(suite.report.get_slack_report()),
            channel=suite.slack_notify.channel,
        )


@click.command(help="Run a given suite.")
@click.option(
    "--suite",
    help="path to the suite file",
    type=click.Path(exists=True),
    required=True,
)
@click.pass_context
def run(ctx: click.Context, suite: str):
    s = Path(suite)
    asyncio.run(main(s))


if __name__ == "__main__":
    run()



import asyncio
from asyncio import Queue
from dataclasses import dataclass
import logging
from pathlib import Path
from time import time
from typing import Dict

from openapi_client import SimulationsApi, ApiClient, Simulation
from openapi_client.models import SimulationStatus
from rich.progress import (
    BarColumn,
    TextColumn,
    Progress,
)

from .base import class_repr, evaluate_lazy
from .base import SimulationChildWrapper
from cmd_client.console import console

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 30


class WaitForTimeout(Exception):
    pass


def match(websocket_instance, data: Dict):
    """
    Recursively try to match a dict with the websocket_instance
    (openapi model instance)
    """
    if websocket_instance is None:
        return True

    mismatch = False

    for key, value in data.items():

        if isinstance(value, dict):
            mismatch = match(getattr(websocket_instance, key, None), value)
        else:
            # Support both object and dict lookups
            if isinstance(websocket_instance, dict):
                val = websocket_instance.get(key, None)
            else:
                val = getattr(websocket_instance, key, None)
            # Allow lazy evaluation of lambda functions
            if val != evaluate_lazy(value):
                mismatch = True
        if mismatch:
            break
    return mismatch


class WaitForBase(SimulationChildWrapper):
    api_class = SimulationsApi
    model = None  # Override this
    scenario_name = ""  # Override this
    websocket_event_type = ""  # Override this

    def __init__(
        self,
        data: Dict,
        api_client: ApiClient,
        simulation: Simulation,
        base_path: Path = None,
    ):
        """Set the timeout to -1 to wait infinitely for the event"""
        if "timeout" in data:
            self.timeout = data.pop("timeout")
        else:
            self.timeout = DEFAULT_TIMEOUT
        super().__init__(data, api_client, simulation, base_path)

    def get_instance(self, data: Dict):
        instance = None
        event_type = data.get("type", None)
        event_data = data.get("data", {})

        if event_type == self.websocket_event_type:
            instance = self.model(**event_data)
        return instance

    def matches(self, websocket_instance) -> bool:
        mismatch = match(websocket_instance, self._data)
        return not mismatch

    async def execute(self, queue: Queue):
        description = class_repr(self._data, self.model)
        console.rule(f"[bold] Waiting for {description}", style="gold3")

        progress = Progress(
            TextColumn("[bold blue]Calculating running...", justify="right"),
            BarColumn(bar_width=25),
            TextColumn("Timeout: {task.fields[timeout]}"),
            "/",
            "{task.fields[remaining]} secs",
            "•",
            "current t: {task.fields[t]}",
        )
        start = time()
        if self.timeout == -1:
            remaining_time = None
        else:
            remaining_time = float(self.timeout)
        waiting = True
        sim_time = 0
        with progress:
            task = progress.add_task(
                "busy",
                total=1,
                start=False,
                timeout=self.timeout,
                remaining=remaining_time,
                t="",
            )
            while waiting:
                try:
                    data = await asyncio.wait_for(
                        queue.get(), timeout=remaining_time
                    )
                    try:
                        sim_time = data["sim_time"]
                    except (KeyError, TypeError):
                        pass
                    progress.update(
                        task, remaining=int(remaining_time), t=sim_time
                    )
                except asyncio.TimeoutError:
                    raise WaitForTimeout(f"Timeout error for: {description}")
                instance = self.get_instance(data)
                if instance is not None:
                    if self.matches(instance):
                        progress.start_task(task)
                        progress.update(task, advance=1)
                        waiting = False

                if (
                    data.get("type") == "websocket"
                    and data.get("status") == "closing"
                ):
                    raise WaitForTimeout(data.get("reason"))

                if waiting and remaining_time is not None:
                    remaining_time = float(self.timeout) - (time() - start)
                    if remaining_time <= 0:
                        raise WaitForTimeout(
                            f"Timeout error for: {description}"
                        )
        console.print(f":heavy_check_mark: Received {self}!")
        console.print("• Resuming scenario...")
        return instance

    def __repr__(self):
        return class_repr(self._data, self.model)


class WaitForModel(WaitForBase):
    websocket_event_type = "event"

    def get_instance(self, data: Dict):
        instance = None
        event_type = data.get("type", None)

        if event_type == self.websocket_event_type:
            model_data = data.get("data", {})
            if model_data.get("model", None) == self.websocket_model_name:
                instance = self.model(**model_data.get("data", {}))
        return instance

    @property
    def websocket_model_name(self):
        return self.model.__name__


class WaitForStatusWrapper(WaitForModel):
    api_class = SimulationsApi
    model = SimulationStatus
    scenario_name = "waitforstatus"
    websocket_event_type = "status"


class WaitForProcessedFileWrapper(WaitForModel):
    pass


def match_validated_event(websocket_instance, event: WaitForBase) -> bool:
    if (
        isinstance(websocket_instance, event.model)
        and websocket_instance.id == event.instance.id
    ):
        state = websocket_instance.to_dict().get("state")
        if state == "processing":
            return False
        elif state == "invalid":
            raise ValueError(
                f"Invalid {event.model}: {websocket_instance.state_detail}"
            )
        elif state == "valid":
            return True

    return False


class WaitForEventValidation(WaitForModel):
    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


@dataclass
class Time:
    time: int = None
    load: float = None
    simulation_id: str = None


class WaitForTimeWrapper(WaitForBase):
    api_class = SimulationsApi
    model = Time
    scenario_name = "waitfortime"
    websocket_event_type = "time"

    def matches(self, websocket_instance):
        if getattr(websocket_instance, "time") >= self._data["time"]:
            return True

        return False


WRAPPERS = [
    WaitForStatusWrapper,
    WaitForTimeWrapper,
]



from slack import WebClient
from slack.errors import SlackApiError


class SlackNotifier:
    def __init__(self, token):
        self.client = WebClient(token=token)

    def send_message(self, message: str, channel: str):
        try:
            self.client.chat_postMessage(channel=channel, text=message)
        except SlackApiError as e:
            # You will get a SlackApiError if "ok" is False
            assert e.response["ok"] is False
            assert e.response["error"]
            raise e



import asyncio
import json
from typing import Optional
import logging
import pkg_resources

import websockets
from websockets.http import Headers
from cmd_client.commands.settings import WebSocketSettings
from cmd_client.console import console

logger = logging.getLogger(__name__)


class WebsocketClient(object):
    def __init__(self, settings: WebSocketSettings):
        self.settings = settings
        self.websocket = None
        self.do_listen = True
        self.queue = asyncio.Queue()
        self._connected = False

    def get_queue(self):
        return self.queue

    async def is_connected(self):
        while self._connected is False:
            await asyncio.sleep(0.5)

    @property
    def user_agent(self):
        return {
            "user-agent": f"3Di-cmd-client/{pkg_resources.get_distribution('3Di-cmd-client').version}"
        }

    async def listen(self, endpoint_uri: str):
        uri = f"{self.settings.proto}://{self.settings.host}/{self.settings.api_version}/{endpoint_uri}"
        console.print(f"• Trying to connect to {uri} now...")
        headers = Headers(authorization=f"{self.settings.token}")
        headers.update(**self.user_agent)
        sim_time: Optional[int] = None
        async with websockets.connect(uri, extra_headers=headers) as websocket:
            console.print(f"•  Connected to {uri}")
            self._connected = True
            async for message in websocket:
                try:
                    message = json.loads(message)
                    try:
                        sim_time = message["data"]["time"]
                    except (KeyError, TypeError):
                        pass
                    if sim_time is not None:
                        message["sim_time"] = sim_time
                    await self.queue.put(message)
                except websockets.exceptions.ConnectionClosedOK:
                    self.do_listen = False
        console.print("• Websocket connection closed")

    async def close(self):
        self.do_listen = False
        if self.websocket:
            await self.websocket.close()



import asyncio
from dataclasses import dataclass, field
from dateutil import parser
from enum import Enum
import json
from typing import Dict, List
from typing import Optional
from typing import NoReturn
from functools import cached_property

from rich.progress import (
    BarColumn,
    TextColumn,
    TimeRemainingColumn,
    Progress,
    Task,
)

from cmd_client.commands.settings import (
    get_settings,
)
from cmd_client.websockets.clients import WebsocketClient
from cmd_client.console import console

MAX_LEN_NAME = 30


class StatusColors:
    starting = "gold3"
    initialized = ""
    finished = "green"
    postprocessing = "royal_blue1"
    crashed = "red"
    ended = "dodger_blue1"


class MessageTypes(Enum):
    initial = "active-simulations"
    progress = "progress"
    status = "status"
    new = "active-simulation"


progress = Progress(
    TextColumn(
        "[bold blue]ID {task.fields[sim_id]} | {task.fields[status]}",
        justify="left",
    ),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    TextColumn("[bold cyan]{task.fields[name]}", justify="left"),
    "•",
    TextColumn("[bold]{task.fields[user]}", justify="left"),
    "•",
    TextColumn(
        "[bold dark_goldenrod]{task.fields[organisation]}", justify="left"
    ),
    "•",
    TimeRemainingColumn(),
    console=console,
    auto_refresh=False,
    # transient=True
)


@dataclass
class ProgressInfo:
    # defaults are used for the header,
    # instance overrides these values
    sim_id: str = "ID"
    progress: str = "PROGRESS"
    name: str = "NAME"
    user: str = "USER"
    organisation: str = "ORGANISATION"
    time_reamining: str = "TIME REMAINING"

    @classmethod
    def header(cls):
        return f"Currently running simulations [{cls.sim_id} • {cls.progress} • {cls.name} • {cls.user} • {cls.organisation} • {cls.time_reamining}]"  # noqa


@dataclass
class ProgressTask(ProgressInfo):
    """
    every progress bar is managed through a rich.progress.Task.
    This class holds all task related info, plus the actual
    rich.progress.Task instance called 'task'
    """

    description: str = "running"
    start: bool = field(init=False)
    total: int = 100
    status: str = ""
    task: Task = None
    is_live: bool = False

    _kwarg_fields = {
        "description",
        "sim_id",
        "name",
        "user",
        "organisation",
        "start",
        "total",
        "status",
    }

    def __post_init__(self):
        self.start = True if self.progress else False

    @property
    def kwargs(self) -> Dict:
        d = {
            field_name: getattr(self, field_name)
            for field_name in self._kwarg_fields
        }
        return {k: v for k, v in d.items() if not k.startswith("_")}


class ActiveSimulations:

    STATUS_COLORS = {
        "starting": "gold3",
        "initialized": "",
        "finished": "green",
        "postprocessing": "royal_blue1",
        "crashed": "red",
        "ended": "dodger_blue1",
    }

    def __init__(self, endpoint: str):
        self.endpoint = endpoint  # local | stag | prod
        # local cache {simulation_id: ProgressTask}
        self.tasks = {}

    @cached_property
    def websocket_client(self):
        settings = get_settings(self.endpoint)
        return WebsocketClient(settings=settings.websocket_settings)

    async def add_progress_bar(
        self, progress_task: ProgressTask, progress: Progress
    ) -> None:
        """creates a new progress task and adds a ProgressTask instance to the internal registry"""
        task_id = progress.add_task(**progress_task.kwargs)
        if progress_task.progress:
            progress.update(task_id, advance=int(progress_task.progress))
        task = [t for t in progress.tasks if t.id == task_id][0]
        progress_task.task = task
        self.tasks[int(progress_task.sim_id)] = progress_task

    async def advance_progress(self, data: Dict, progress: Progress) -> None:
        """advances the progress bar for the given tasks by the current simulation progress"""
        simulation_id = data["data"]["simulation_id"]
        sim_progress = data["data"]["progress"]
        progress_task = self.tasks.get(int(simulation_id))
        if not progress_task.task.started:
            progress.start_task(progress_task.task.id)
        if int(sim_progress) <= progress_task.task.percentage:
            return
        advance_by = int(sim_progress) - int(progress_task.task.percentage)
        progress.update(progress_task.task.id, advance=advance_by)
        # auto_refresh is False so do this manually
        progress.refresh()

    @staticmethod
    def is_live_finished(status, progress_task: ProgressTask) -> bool:
        return all(
            (
                status == "crashed",
                progress_task.is_live,
                progress_task.status in ("ended", "postprocessing"),
            )
        )

    async def update_status(self, data: Dict, progress: Progress):
        status = data["data"]["status"]
        simulation_id = data["data"]["simulation_id"]
        progress_task = self.tasks.get(int(simulation_id))
        if self.is_live_finished(status, progress_task):
            status = "finished"
        status_txt = status
        color = self.STATUS_COLORS.get(status)
        if color:
            status_txt = f"[bold {color}]{status_txt}[/bold {color}]"
        update_kwargs = {"status": status_txt}
        advance_by = (
            100 if status in ("finished", "ended", "postprocessing") else None
        )
        if advance_by and not progress_task.task.started:
            progress.start_task(progress_task.task.id)
            update_kwargs.update({"advance": advance_by})
        progress.update(progress_task.task.id, **update_kwargs)
        # auto_refresh is False so do this manually
        progress.refresh()
        progress_task.status = status

    async def run_monitor(self) -> NoReturn:

        asyncio.ensure_future(
            self.websocket_client.listen("active-simulations/")
        )
        q = self.websocket_client.get_queue()
        try:
            await asyncio.wait_for(
                self.websocket_client.is_connected(), timeout=5
            )
        except asyncio.TimeoutError:
            console.print("Could not establish WS connection", style="error")
            return
        with progress:
            progress.console.rule(ProgressInfo.header())
            while True:
                data = await q.get()
                message_type = await self.get_msg_type(data)
                if not message_type:
                    continue
                if message_type == MessageTypes.initial:
                    progress_tasks = await self.get_initial_progress_tasks(
                        data
                    )
                    for progress_task in progress_tasks:
                        await self.add_progress_bar(progress_task, progress)
                elif message_type == MessageTypes.progress:
                    await self.advance_progress(data, progress)
                elif message_type == MessageTypes.status:
                    await self.update_status(data, progress)
                elif message_type == MessageTypes.new:
                    progress_task = await self.get_new_progress_task(data)
                    await self.add_progress_bar(progress_task, progress)

    @staticmethod
    async def get_msg_type(data) -> Optional[MessageTypes]:
        try:
            return MessageTypes(data["type"])
        except (KeyError, ValueError):
            return

    @staticmethod
    async def get_new_progress_task(data):
        """
        message from "active-simulations" URI
        type declaration "active-simulation"
        """
        data = data["data"]
        simulation_id = list(data.keys())[0]
        data_json = list(data.values())[0]
        return await ActiveSimulations._get_progress_task(
            simulation_id, data_json
        )

    @staticmethod
    async def _get_progress_task(
        simulation_id, simulation_details
    ) -> ProgressTask:

        details = json.loads(simulation_details)
        if len(details["name"]) > MAX_LEN_NAME:
            name = f"{details['name'][:MAX_LEN_NAME]}..."
        else:
            name = details["name"]

        is_live = int(details["duration"]) == 3153600000
        if is_live:
            dt = parser.parse(details["date_created"])
            name = f"{name} | {dt.strftime('%Y-%m-%d %H:%M')} [bold red][LIVE][/bold red]"
        sim_progress = details["progress"]
        return ProgressTask(
            sim_id=simulation_id,
            name=name,
            user=details["user_name"],
            organisation=details["organisation_name"],
            progress=sim_progress,
            status=details["status"],
            is_live=is_live,
        )

    @staticmethod
    async def get_initial_progress_tasks(data: Dict) -> List[ProgressTask]:
        progress_tasks = []
        for simulation_id, json_details in data["data"].items():
            task_info = await ActiveSimulations._get_progress_task(
                simulation_id, json_details
            )
            progress_tasks.append(task_info)
        return progress_tasks



import yaml
from jinja2 import Environment
from uuid import uuid4
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

from threedi_api_client.threedi_api_client import ThreediApiClient

from cmd_client.models import WRAPPERS
from cmd_client.models.scenario import Scenario
from cmd_client.commands.settings import WebSocketSettings


class ScenarioParser:
    """
    Parses a YAML file into a Scenario object
    """

    def __init__(self, filepath: str, context: Dict):
        """
        :param filepath: The filepath to the YAML file
        :param context: The context to use for rendering the YAML (Jinja2) file
        """
        self.filepath = Path(filepath)
        self.context = context

    def get_context(self) -> Dict:
        """
        Get default context
        """
        context = {
            "simulation_name": "run_" + uuid4().hex,
            "datetime_now": datetime.utcnow().isoformat(),
        }
        context.update(self.context)
        return context

    def _render_template(self) -> str:
        """
        Render the template using the given context
        """
        with open(self.filepath.as_posix()) as f:
            env = Environment(extensions=["jinja2_time.TimeExtension"])
            env.datetime_format = "%Y-%m-%dT%H:%M:%S"
            data = env.from_string(f.read()).render(self.get_context())
        return data

    def parse(
        self,
        threedi_api_client: ThreediApiClient,
        websocket_settings: WebSocketSettings,
        base_path: Optional[Path] = None,
    ) -> Scenario:
        """
        Parse the YAML file.

        :param threedi_api_client: Injected into the Scenario,
                                   allowing to execute API calls.

        :returns: Scenario instance
        """
        template = self._render_template()
        data = yaml.load(template, Loader=yaml.FullLoader)
        return Scenario(
            data=data,
            threedi_api_client=threedi_api_client,
            wrappers=WRAPPERS,
            websocket_settings=websocket_settings,
            base_path=base_path,
        )



import logging
from .base import SimulationChildWrapper
from openapi_client import SimulationsApi
from openapi_client.models import Action
from .waitfor import WaitForStatusWrapper, WaitForTimeWrapper

from cmd_client.console import console
from rich.live import Live

logger = logging.getLogger(__name__)


class ActionWrapper(SimulationChildWrapper):
    api_class = SimulationsApi
    model = Action
    api_path: str = "actions"
    scenario_name = "action"

    def create(self):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = getattr(self.api, f"{self.base_path}{self.api_path}_create")

            data = self.instance.to_dict()
            data = {
                key: item for key, item in data.items() if item is not None
            }

            res = func(self.simulation.id, data)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    @property
    def extra_steps(self):
        extra_steps = []
        name = self.instance.name
        if name == "initialize":
            data = {"name": "initialized", "paused": True}
        elif name == "start":
            data = {"name": "initialized", "paused": False}
        elif name == "pause":
            data = {"name": "initialized", "paused": True}
        elif name == "shutdown":
            data = {"name": "finished"}
        elif name == "queue":
            msg = "'queue' step is not yet implemented"
            raise NotImplementedError(msg)
        else:
            msg = f"Unknown name {name}"
            raise ValueError(msg)

        extra = {}
        if "waitfor_timeout" in self._data:
            extra["timeout"] = self._data["waitfor_timeout"]

        wait_for_status = WaitForStatusWrapper(
            data={**data, **extra},
            api_client=self._api_client,
            simulation=self.simulation,
        )
        extra_steps.append(wait_for_status)

        if self.instance.duration is not None:
            extra_steps.append(
                WaitForStatusWrapper(
                    data={"name": "initialized", "paused": True, **extra},
                    api_client=self._api_client,
                    simulation=self.simulation,
                )
            )
            extra_steps.append(
                WaitForTimeWrapper(
                    data={"time": self.instance.duration, **extra},
                    api_client=self._api_client,
                    simulation=self.simulation,
                )
            )
        return extra_steps



from .base import ModelWrapper
from openapi_client.api import SimulationsApi
from openapi_client.models import Simulation


class SimulationWrapper(ModelWrapper):
    api_class = SimulationsApi
    model = Simulation
    api_path = "simulations"
    scenario_name = "simulation"



from enum import Enum


class ExitCodes(Enum):
    OK = 0
    # CONFIG_ERROR = 4
    LOADING_SUITE_ERROR = 5
    SCENARIO_CONFIG_ERROR = 6
    RUN_SCENARIO_ERROR = 7
    CONNECTION_ERROR = 8
    AUTHENTICATION_FAILED = 9

    # FILE_NOT_FOUND = 6


class LoadCredentialsError(BaseException):
    pass


class LoadSuiteError(BaseException):
    pass


class LoadScenarioError(BaseException):
    pass



import asyncio
import logging
import pathlib
import sys
from datetime import datetime

import click
from openapi_client import Configuration, ApiException
from openapi_client.api.simulations_api import SimulationsApi
from openapi_client.api.organisations_api import OrganisationsApi
from openapi_client.api.threedimodels_api import ThreedimodelsApi
from rich.table import Table
from rich import box
from rich.prompt import Confirm
from rich.prompt import Prompt
from rich.padding import Padding
from rich.panel import Panel

from cmd_client.commands.settings import (
    Settings,
    EndpointOption,
    refresh_api_key,
    get_settings,
)
from cmd_client.commands.utils import download_files
from cmd_client.parser import ScenarioParser
from cmd_client.console import console
from cmd_client.errors import ExitCodes
from cmd_client.models.errors import ApiModelError
from cmd_client.models.scenario import ResolveError
from cmd_client.models.scenario import FailedStep
from cmd_client.errors import LoadScenarioError


# Logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter(f" > %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

DEFAULT_PAGER_SIZE = 20


@click.group()
@click.option(
    "--endpoint",
    type=click.Choice(EndpointOption.__members__.keys()),
    default=EndpointOption.production.name,
    help="The endpoint where commands are run against (defaults to production --> https://api.3di.live/v3.0). "
    "The other endpoints are only useful for staff members.",
)
@click.pass_context
def cli(ctx: click.Context, endpoint: str):
    settings = get_settings(endpoint)
    ctx.obj = settings
    ctx.call_on_close(Settings.save_settings)


@cli.command()
@click.pass_obj
def auth(settings: Settings):
    """Provide authentication details"""
    try:
        configuration = Configuration(settings.endpoint)
        refresh_api_key(configuration)
        console.print(
            f":unlock: Authenticated as {settings.username}", style="success"
        )
    except ApiException as e:
        console.print(
            f":lock: Failed to authenticate: {e.reason}", style="warning"
        )
        sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)


@cli.command()
@click.pass_obj
def models(settings: Settings):
    """List available threedimodels"""
    threedi_models_api = ThreedimodelsApi(settings.api_client)
    threedi_models = threedi_models_api.threedimodels_list()

    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column("Id", width=5)
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )
    table.add_column("Revision", justify="left")
    table.add_column("Inp success", justify="left")

    remaining = threedi_models.count
    limit = min(DEFAULT_PAGER_SIZE, remaining)

    while remaining == threedi_models.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        threedi_models = threedi_models_api.threedimodels_list(
            limit=limit, offset=remaining - limit
        )
        for i, model in enumerate(threedi_models.results):
            if model.inp_success is True:
                txt = f":heavy_check_mark: [bold green]{model.inp_success}[/bold green] "
            else:
                txt = f"[bold red]{model.inp_success}[/bold red]"
            table.add_row(
                f"{model.id}", f"{model.name}", f"{model.revision_hash}", txt
            )
        remaining -= limit
        limit = min(remaining, DEFAULT_PAGER_SIZE)
        console.print(table)


@cli.command()
@click.pass_obj
def organisations(settings: Settings):
    """List available organisations"""
    organisations_api = OrganisationsApi(settings.api_client)
    organisations = organisations_api.organisations_list()
    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column("Unique Id", width=15)
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )

    remaining = organisations.count
    limit = min(DEFAULT_PAGER_SIZE, remaining)

    while remaining == organisations.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        organisations = organisations_api.organisations_list(
            limit=limit, offset=remaining - limit
        )
        for i, org in enumerate(organisations.results):
            table.add_row(f"{org.unique_id}", f"{org.name}")
        remaining -= limit
        limit = min(remaining, DEFAULT_PAGER_SIZE)
        console.print(table)


@cli.command()
@click.pass_obj
def simulations(settings: Settings):
    """List simulations"""
    simulations_api = SimulationsApi(settings.api_client)
    simulations = simulations_api.simulations_list()

    table = Table(
        show_header=True,
        box=box.HORIZONTALS,
        show_lines=False,
        width=(console.width * 80) / 100,
    )
    table.add_column(
        "Id", width=5
    )  # , style="dark_green", header_style="bold dark_green")
    table.add_column("Name", width=20, justify="left")
    table.add_column("Status", justify="left")

    remaining = simulations.count
    limit = min(DEFAULT_PAGER_SIZE, remaining)
    while remaining == simulations.count or (
        remaining > 0 and Confirm.ask("Show more?")
    ):
        simulations = simulations_api.simulations_list(
            limit=limit, offset=remaining - limit
        )
        for i, simulation in enumerate(reversed(simulations.results)):
            status = simulations_api.simulations_status_list(simulation.id)
            if status.name == "finished":
                txt = f":heavy_check_mark: [bold green]{status.name}[/bold green] "
            elif status.name == "created":
                txt = f"[dim grey27]{status.name}[/dim grey27] "
            elif status.name == "initialized":
                txt = f"[bold dark_violet]{status.name}[/bold dark_violet] "
            else:
                txt = f"[bold red]{status.name}[/bold red]"
            table.add_row(f"{simulation.id}", f"{simulation.name}", txt)
        console.print(table)
        remaining -= limit
        limit = min(remaining, DEFAULT_PAGER_SIZE)


@cli.command()
@click.pass_obj
def scenarios(settings: Settings):
    """List local scenarios"""
    table = Table(show_header=True, box=box.HORIZONTALS, show_lines=True)
    table.add_column(
        "Id", width=5
    )  # , style="dark_green", header_style="bold dark_green")
    table.add_column(
        "Name",
        width=20,
        justify="left",
        style="bold cyan",
        header_style="bold cyan",
    )
    table.add_column("Description", justify="left", no_wrap=False)
    table.add_column(
        "Caution", justify="left", style="orange3", header_style="bold orange3"
    )
    table.add_column(
        "yaml", justify="left", style="dim blue", header_style="bold blue"
    )

    for i, scenario in enumerate(settings.scenarios, start=0):
        if isinstance(scenario, dict):
            s = ""
            if scenario.get("known_constraints"):
                for k, v in scenario["known_constraints"].items():
                    s += f"{k}: {v} \n"
            table.add_row(
                f"{i}",
                f"{scenario['name']}",
                f"{scenario['description']}",
                f"{s}",
                f"{scenario['file'].stem}",
            )
    console.print(table)


@cli.command()
@click.option(
    "--organisation",
    type=str,
    help="Unique-id of the organisation to set as default",
)
@click.option(
    "--result-folder",
    type=click.Path(dir_okay=True, writable=True, resolve_path=True),
)
@click.pass_obj
@click.pass_context
def settings(ctx, settings: Settings, organisation, result_folder):
    """Set default settings"""
    console.rule(
        ":wrench: Configuring defaults for the 3Di cli", style="bold blue"
    )
    console.print(Padding("", (1, 0)))
    ctx.invoke(auth)
    if organisation is None:
        console.rule("Please choose an organisation", style="bold blue")
        if (
            settings.organisation_uuid
            and Confirm.ask(
                f"Change current organisation {settings.organisation_uuid}?"
            )
            or not settings.organisation_uuid
        ):
            ctx.invoke(organisations)
            console.rule(":pencil2:", style="bold blue")
            organisation = Prompt.ask(
                "Set default organisation. UNIQUE_ID",
                default=settings.organisation_uuid,
            )
        else:
            organisation = settings.organisation_uuid
    if not result_folder:
        result_folder = click.prompt(
            "Set default result folder",
            default=settings.result_folder or "./results",
            type=click.Path(dir_okay=True, writable=True, resolve_path=True),
        )

    settings.organisation_uuid = organisation
    settings.result_folder = result_folder
    console.print(
        f":heavy_check_mark: Default settings are saved in {settings.config_file}"
    )


@cli.command(help="Run a given scenario")
@click.option(
    "--scenario",
    type=int,
    help="The ID of the scenario, as returned by the 'scenarios' command",
)
@click.option(
    "--model",
    type=int,
    help="The ID of the model you want to use, as returned by the 'models' command",
)
@click.option(
    "--organisation",
    type=str,
    help="The UUID of the organisation you want to run the scenario for, "
    "as returned by the 'organisations' command",
)
@click.pass_obj
@click.pass_context
def run(ctx, settings: Settings, scenario: int, model: int, organisation):
    """Run a scenario"""
    if scenario is None:
        ctx.invoke(scenarios)
        scenario = click.prompt(
            "Which scenario do you want to run? ID", type=int
        )

    if model is None:
        ctx.invoke(models)
        model = click.prompt("Which model do you want to run? ID", type=int)

    if not organisation:
        organisation = settings.organisation_uuid

    scenario_to_run = settings.scenarios[scenario]["file"]
    name = settings.scenarios[scenario]["name"]
    context = {
        "threedimodel_id": model,
        "organisation_uuid": organisation,
        "simulation_name": name,
        "datetime_now": datetime.utcnow().isoformat(),
    }
    parser = ScenarioParser(scenario_to_run, context)
    try:
        scenario = parser.parse(
            settings.api_client, settings.websocket_settings
        )
    except (ResolveError, ApiModelError, LoadScenarioError) as err:
        console.print(f":collision: {err}", style="error")
        sys.exit(ExitCodes.SCENARIO_CONFIG_ERROR.value)

    console.rule(f"Loading scenario {name}", style="bold blue")
    scenario.simulation.save()
    console.print(f":link: URL: {scenario.simulation.instance.url}")
    import websockets

    try:
        console.rule(f"Starting scenario run...", style="bold blue")
        asyncio.run(scenario.execute())
    except KeyboardInterrupt:
        pass
    except FailedStep as err:
        console.print(f"{err}", style="error")
        sys.exit(ExitCodes.RUN_SCENARIO_ERROR.value)
    except websockets.exceptions.InvalidStatusCode as err:
        console.print(f"{err}", style="error")
        sys.exit(ExitCodes.CONNECTION_ERROR.value)
    else:
        success_panel = Panel(
            f"Run for scenario {name} successful",
            expand=True,
            box=box.DOUBLE,
            border_style="bold spring_green4",
            title=f":sparkles: Finished :sparkles:",
        )
        console.print(success_panel, justify="center")


@cli.command(help="Download results of a simulation")
@click.option("--simulation", type=int, help="ID of the simulation")
@click.option(
    "--folder",
    type=click.Path(dir_okay=True, writable=True, resolve_path=True),
    help="Absolute path to the where the files will be stored",
)
@click.pass_obj
@click.pass_context
def results(ctx, settings: Settings, simulation: int, folder):
    """Download results of a simulation"""
    console.rule(
        ":arrow_heading_down:  Download simulation results", style="bold blue"
    )
    console.print(Padding("", (1, 0)))

    if simulation is None:
        console.rule("Please choose a simulation", style="bold blue")
        ctx.invoke(simulations)
        simulation = click.prompt(
            "Which simulation do you want download? ID", type=int
        )
    if not folder:
        folder = click.prompt(
            "Where do you want to store the results files?",
            default=f"{settings.result_folder}/simulation-{simulation}",
            type=click.Path(dir_okay=True, writable=True, resolve_path=True),
        )

    output_folder = pathlib.Path(folder)
    try:
        output_folder.mkdir(parents=True)
    except FileExistsError:
        click.confirm(
            "Output folder already exists, we might override files in the folder. "
            "Do you want to continue?",
            abort=True,
        )

    simulations_api = SimulationsApi(settings.api_client)
    threedimodels_api = ThreedimodelsApi(settings.api_client)

    simulation = simulations_api.simulations_read(id=simulation)
    threedi_model_id = simulation.threedimodel_id

    gridadmin_download = threedimodels_api.threedimodels_gridadmin_download(
        threedi_model_id
    )
    f = [gridadmin_download]

    result_files = simulations_api.simulations_results_files_list(
        simulation.id
    )
    for result in result_files.results:
        if result.file.state in ["error", "removed"]:
            console.print(
                f"{result.filename} is in state {result.file.state} and will be skipped.",
                style="warning",
            )
            continue

        result_download = simulations_api.simulations_results_files_download(
            id=result.id, simulation_pk=simulation.id
        )
        f.append(result_download)
    success = download_files(f, output_folder)
    if success:
        console.print(
            ":heavy_check_mark: Finished downloading results", style="success"
        )
    else:
        console.print(
            ":collision: Not all files could be downloaded", style="error"
        )


if __name__ == "__main__":
    cli()



import json
import logging
from asyncio import Queue
from inspect import isfunction
from typing import Dict, List
from openapi_client import ApiClient
from openapi_client.models import Simulation
from openapi_client.api import SimulationsApi
from pathlib import Path
import requests
from cmd_client.console import console
from cmd_client.models.errors import ApiModelError
from rich.live import Live

logger = logging.getLogger(__name__)


def evaluate_lazy(value):
    if isfunction(value):
        return value()
    return value


def class_repr(data, model):
    params = ", ".join(
        [
            f"{key}={json.dumps(str(evaluate_lazy(value)))}"
            for key, value in data.items()
        ]
    )
    return f"{model.__name__}({params})"


class BaseWrapper:
    def __init__(
        self, data: Dict, api_client: ApiClient, base_path: Path = None
    ):
        self._api_client = api_client
        self._data = data
        self._base_path = base_path

    async def execute(self, queue: Queue):
        pass

    @property
    def extra_steps(self):
        return []


BLACKLIST = {"waitfor_timeout"}


class ModelWrapper(BaseWrapper):
    api_class: object = None
    api_path: str = None
    instance: object = None
    _internal_state = None
    model: object = None
    scenario_name = None

    def __init__(
        self, data: Dict, api_client: ApiClient, base_path: Path = None
    ):
        super().__init__(data, api_client, base_path)
        self.initialize_instance(data)
        self.api = self.api_class(api_client)  # noqa

    def initialize_instance(self, data: Dict):
        """raises ApiModelError if initializing the instance fails"""
        new_data = {
            key: value
            for key, value in data.items()
            if not key.startswith("__") and key not in BLACKLIST
        }
        try:
            self.instance = self.model(**new_data)
        except ValueError as err:
            msg = f"Initializing {self.model.__name__} failed: {err}"
            raise ApiModelError(msg)

    async def execute(self, queue):
        # By default save()
        self.save()
        for step in self.extra_steps:
            await step.execute(queue)

    def create(self):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = getattr(self.api, f"{self.api_path}_create")
            res = func(self.instance)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model.__name__}...")
            func = getattr(self.api, f"{self.api_path}_partial_update")
            res = func(self.instance.id, data)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Patched {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def delete(self):
        with Live(console=console) as live:
            live.update(f"Patching {self.model.__name__}...")
            func = getattr(self.api, f"{self.api_path}_delete")
            res = func(self.instance.id)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Deleted {self.model.__name__}"
            )
        return res

    def save(self):
        res = None
        id = getattr(self.instance, "id", None)

        if id is None:
            # Create
            res = self.create()
        else:
            # Update (use patch by default)
            old_set = set(self._internal_state.to_dict().items())
            new_set = set(self.instance.to_dict().items())

            diff = new_set - old_set

            if diff != set():
                # There is something we can try to update
                res = self.patch(dict(diff))

        if res is not None:
            # Update values of self to match what has been returned
            for key, value in res.to_dict().items():
                if hasattr(self.instance, key):
                    setattr(self.instance, key, value)

            self._internal_state = res

    def __repr__(self):
        return class_repr(self._data, self.model)


class SimulationWrapper(ModelWrapper):
    api_class = SimulationsApi
    model = Simulation
    api_path = "simulations"


class SimulationChildWrapper(ModelWrapper):
    simulation = None
    base_path = "simulations_"

    def __init__(
        self,
        data: Dict,
        api_client: ApiClient,
        simulation: Simulation,
        base_path: Path = None,
    ):
        self.simulation = simulation
        super().__init__(data, api_client, base_path)

    @property
    def instance_name(self):
        return self.scenario_name

    def create(self, *args, **kwargs):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = getattr(self.api, f"{self.base_path}{self.api_path}_create")
            res = func(self.simulation.id, self.instance, *args, **kwargs)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model}...")
            func = getattr(
                self.api, f"{self.base_path}{self.api_path}_partial_update"
            )
            res = func(self.instance.id, self.simulation.id, data)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Patched {self.model}"
            )
        return res

    def delete(self):
        func = getattr(self.api, f"{self.base_path}{self.api_path}_delete")
        res = func(self.instance.id, self.simulation.id)
        logger.debug(
            "Deleted %s with id: %s", self.instance_name, self.instance.id
        )
        return res


class EventWrapper(SimulationChildWrapper):
    base_path = "simulations_events_"

    def save(self):
        super().save()
        if not hasattr(self, "filepath") or not self.filepath:
            return
        filepath = Path(self.filepath)
        if self._base_path:
            filepath = self._base_path / filepath
        fp = (filepath).resolve()
        # Upload file
        with fp.open("rb") as f:
            res = requests.put(self.instance.put_url, data=f)
            assert res.status_code == 200


class InitialWrapper(SimulationChildWrapper):
    base_path = "simulations_initial_"



from typing import Dict
from pathlib import Path

from openapi_client import SimulationsApi, Upload
from openapi_client.models import FileBoundaryCondition

from .base import EventWrapper
from .waitfor import WaitForEventValidation


class WaitForFileBoundaryConditionsWrapper(WaitForEventValidation):
    api_class = SimulationsApi
    model = FileBoundaryCondition
    scenario_name = "waitforfileboundaryconditions"


class FileBoundaryConditionsWrapper(EventWrapper):
    api_class = SimulationsApi
    model = Upload
    api_path: str = "boundaryconditions_file"
    scenario_name = "fileboundaryconditions"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        file_boundary_conditions = (
            self.api.simulations_events_boundaryconditions_file_list(
                simulation_pk=self.simulation.id
            )
        )
        for file_boundaryconditions in file_boundary_conditions.results:
            if file_boundaryconditions.file:
                file_boundaryconditions.file.filename == self.instance.filename
                bulk_event_instance = file_boundaryconditions
                break
        assert bulk_event_instance is not None

        wait_for_validation = WaitForFileBoundaryConditionsWrapper(
            data=bulk_event_instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


WRAPPERS = [
    WaitForFileBoundaryConditionsWrapper,
    FileBoundaryConditionsWrapper,
]



import logging
from .base import EventWrapper
from openapi_client import SimulationsApi
from openapi_client.models import TimedStructureControl

from cmd_client.models.waitfor import (
    WaitForEventValidation,
    match_validated_event,
)

logger = logging.getLogger(__name__)


class WaitForStructureControlWrapper(WaitForEventValidation):
    api_class = SimulationsApi
    model = TimedStructureControl
    scenario_name = "waitforstructurecontrol"

    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


class StructureControlWrapper(EventWrapper):
    api_class = SimulationsApi
    model = TimedStructureControl
    api_path: str = "structure_control_timed"
    scenario_name = "structurecontroltimed"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForStructureControlWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


WRAPPERS = [StructureControlWrapper, WaitForStructureControlWrapper]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


with open('README.md') as readme_file:
    readme = readme_file.read()


with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as req_file:
    requirements = req_file.read().splitlines()

with open('requirements_dev.txt') as dev_req_file:
    setup_requirements = dev_req_file.read().splitlines()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

test_requirements = []


setup(
    author="Jelle Prins",
    author_email='info@nelen-schuurmans.nl',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
        'Topic :: Scientific/Engineering',
    ],
    description="Python 3Di command line client",
    long_description_content_type='text/markdown',
    long_description=readme + '\n\n' + history,
    install_requires=requirements,
    license="MIT license",
    entry_points={
        "console_scripts": [
            "scenario=cmd_client.commands.run_scenario:cli",
            "suite=cmd_client.commands.run_suite:run",
            "active_simulations=cmd_client.commands.active_simulations:cli"
        ]
    },
    include_package_data=True,
    keywords='3Di, client, command line, scenario',
    name='3Di_cmd_client',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
        include=['cmd_client', 'cmd_client.*'],
    ),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nens/3Di-cmd',
    version=find_version('cmd_client', 'version.py'),
    zip_safe=False,
)



import asyncio
from functools import cached_property
from pathlib import Path
import sys
from typing import Dict, List
import logging

from openapi_client.models import Simulation
from openapi_client.api import SimulationsApi
from openapi_client.exceptions import ApiException
from threedi_api_client.threedi_api_client import ThreediApiClient

from .base import SimulationWrapper, ModelWrapper, BaseWrapper
from cmd_client.console import console
from cmd_client.websockets.clients import WebsocketClient
from cmd_client.commands.settings import WebSocketSettings
from cmd_client.errors import LoadScenarioError

logger = logging.getLogger(__name__)


class FailedStep(Exception):
    pass


class ResolveError(Exception):
    pass


class Scenario:
    simulation: Simulation = None
    steps: any = []

    def __init__(
        self,
        data: Dict,
        threedi_api_client: ThreediApiClient,
        wrappers: List[BaseWrapper],
        websocket_settings: WebSocketSettings,
        base_path: Path = None,
    ):
        self.wrappers = wrappers
        self.websocket_settings = websocket_settings
        self.threedi_api_client = threedi_api_client
        self._base_path = base_path
        scenario = data.get("scenario")
        if not scenario:
            raise LoadScenarioError("Could not load scenario")
        simulation_data = scenario.get("simulation")
        steps_data = scenario.get("steps", [])
        self.create_simulation(simulation_data)
        self.steps = self.create_steps(steps_data)

    @cached_property
    def websocket_client(self):
        return WebsocketClient(settings=self.websocket_settings)

    def create_simulation(self, simulation_data: Dict):
        self.simulation = SimulationWrapper(
            simulation_data, self.threedi_api_client
        )

    def shutdown_simulation(self):
        console.print("• Requesting simulation shutdown")
        api = SimulationsApi(self.threedi_api_client)

        api.simulations_actions_create(
            self.simulation.instance.id, {"name": "shutdown"}
        )

    def create_steps(self, steps_data: List) -> List[ModelWrapper]:
        """:raises ResolveError if the scenario name can not be found in `steps_data`"""
        steps = []
        if steps_data is None:
            return steps

        for item in steps_data:
            key, data = tuple(item.items())[0]
            found = [x for x in self.wrappers if x.scenario_name == key]

            if not found:
                raise ResolveError(f"Cannot resolve {key} in self.metas")
            wrapper_class = found[0]

            wrapped_instance = wrapper_class(
                data,
                self.threedi_api_client,
                simulation=self.simulation.instance,
                base_path=self._base_path,
            )
            steps.append(wrapped_instance)
        return steps

    async def _run(self):
        from scenario_test_framework.models.waitfor import WaitForTimeout

        queue = self.websocket_client.queue
        try:
            # 'Execute' step items
            for i, item in enumerate(self.steps, start=1):
                await item.execute(queue)
        except ApiException as e:
            msg = f":collision: Failed step {i}: {item} due to APIException: {e.body}"
            raise FailedStep(msg)
        except WaitForTimeout as e:
            msg = f" :collision: Failed step {i}: {item} due to WaitForTimeout: {e}"
            raise FailedStep(msg)
        except Exception as e:
            msg = f" :collision: Failed step {i}: {item} due to: {e}"
            raise FailedStep(msg)
        finally:
            # Always try to cleanup the simulation
            try:
                self.shutdown_simulation()
            except Exception:
                pass

        # Done, close the websocket connection
        try:
            await self.websocket_client.close()
        except Exception as e:
            console.print(e, style="error")

    async def execute(self):
        # Create simulation
        if self.simulation.instance.id is None:
            self.simulation.save()
        endpoint_uri = f"simulations/{self.simulation.instance.id}/?scenario_test_framework=true"
        listen_task = asyncio.create_task(
            self.websocket_client.listen(endpoint_uri)
        )
        try:
            await asyncio.wait_for(
                self.websocket_client.is_connected(), timeout=10
            )
        except asyncio.TimeoutError:
            console.print(
                "Could not establish WebSocket connection within 10 seconds",
                style="error",
            )
            listen_task.cancel()
            await asyncio.gather(listen_task)
            return
        run_task = asyncio.create_task(self._run())
        finished, _ = await asyncio.wait(
            [listen_task, run_task], return_when=asyncio.FIRST_EXCEPTION
        )
        for result in finished:
            if not result:
                continue
            if result.exception():
                raise result.exception()
